// Duelooper Voltage Controlled sample Playback module.
// Copyright (C) 2021 JL Pniewski

 //This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as 
 //published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.

//This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of 
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

//You should have received a copy of the GNU General Public License along with this program. If not, see <http://www.gnu.org/licenses>.

// For Lizzie, Yoshi and Dad.

// 07/08/21 Trig input bug fixed. 

// --------- Import Libraries ---------
#include <SPI.h>
#include <SD.h>

// ---------- Global Variables -----------
volatile int size_of_sample_data;		// In bytes, divide by 2 for no of samples
volatile int playback_start_add;
volatile int abs_spl_read_add;
volatile bool fwd_rev_sw = 0;	// <- rev (1) - fwd (0) ->. Init Fwd as bounce switch may be on and initial Fwd/Rev read would not happen.
volatile bool bounce_sw;	// Bounce up down function switch. On Low, Off High.
volatile bool mode_sw;	// Loop / One-Shot Switch. Loop High, One-Shot Low.
volatile bool restart_sw; // Restart playback Switch. Off high, On Low
volatile int playback_end_add;	// Loop end point
volatile bool has_triggered;	// +ve edge detected flag for trigger input
volatile bool pb_direction;	// current playback direction.
unsigned int spl_add_tbl[4096];		// arrary table non zero crossing points. Due unsigned int 32 bit.
unsigned int zxd_add_tbl[4096];		// array table for zero crossing points. Due unsigned int 32 bit.
unsigned int sample_segment_size;	// Size in samples (Bytes X 2).
bool trig_pol;	// Trigger Polarity
int org_start_pos = 0;	// previous start value. Used to compare with latest AD read for jitter reduction.
int org_end_pos = 0;	// previous end value. Used to compare with latest AD read for jitter reduction.
   
// ---------- Setup -----------
void setup() 
{
	pinMode(LED_BUILTIN, OUTPUT);	// Init Loading / error LED
	pinMode(46, INPUT);	//	Trig/Gate input. Can't use INPUT_PULLUP here as there are triggering issues with a unipolar signal input.
	pinMode(69, INPUT_PULLUP);	//	Trig Polarity select
	pinMode(62, INPUT_PULLUP);	//	Bounce on/off. On Low, Off High.
	pinMode(63, INPUT_PULLUP);	//	FWD/REV Playback. Rev High, Fwd Low
	pinMode(64, INPUT_PULLUP);	//	Zero Crossing Detect on/off. H = off, L = on.
	pinMode(65, INPUT_PULLUP);	//	Restart Switch. On Low, Off High.
	pinMode(48, INPUT_PULLUP);	//	Mode Switch Loop / One-Shot
	
  analogWriteResolution(12); // use full 12bits
  dacc_set_channel_selection(DACC_INTERFACE, 0); // Select DAC 0 for output
  analogWrite(DAC0, 0); 	 // Should be all that is needed to setup DAC0 for direct write.

  //Initialize serial and wait for port to open:
  Serial.begin(57600);
  while (!Serial) {;} // wait for serial port to connect.
  
	// -----------------------------------------------------------------------------------------------------------------
	// PIO register stuff
	unsigned long pio_stat_reg; // pio status register read
    
  // Setup PIO Pins. Port C already set
  
  // Port A PIO Status register. Ram Address Bus Data. Upper bits.
  pio_stat_reg = REG_PIOA_PSR;
  pio_stat_reg |= 0x00063C00; // set all bits as PIO disabling RX1, TX1, RX2, TX2, SDA1, SCL2
  REG_PIOA_PER = pio_stat_reg; // write to register
  
  // Port B Switch and Trigger inputs. B17, B18, B19, B20.
  
  // Port B PIO Status register. 
  pio_stat_reg = REG_PIOB_PSR;
  pio_stat_reg |= 0x001E0000; // B(0000 0000 0001 1110 0000 0000 0000 0000). set all bits as PIO disabling relavent A/D converters 
  REG_PIOB_PER = pio_stat_reg; // write to register

  // Setup Port B inputs
  REG_PIOB_OER = 0x00000000;     // Set Port B bits 17 - 20 to inputs. B(0000 0000 0000 0000 0000 0000 0000 0000)
  REG_PIOB_ODR = 0x001E0000;     // Disable write to Port B bits 17 - 20.     B(0000 0000 0001 1110 0000 0000 0000 0000)
  
  // Setup PIO Clock port B needed for Read operation
  PMC->PMC_PCER0 = PMC->PMC_PCSR0 | 0x01u << ID_PIOB; 
  
  // Port C pins are already set as PIO pins. Lower bits.

  // Port D PIO Status register. Ram Data Bus
  pio_stat_reg = REG_PIOD_PSR;
  pio_stat_reg |= 0x000000FF; // set all bits as PIO disabling TX3 and RX3.
  REG_PIOD_PER = pio_stat_reg; // write to register

  // Setup PIO Clock port D needed for Read operation
  PMC->PMC_PCER0 = PMC->PMC_PCSR0 | 0x01u << ID_PIOD; 

  // Setup Output Enable Registers (Data Direction)
  REG_PIOA_OER =  0x001FFC00;     // Set bits PortA to outputs. B(0000 0000 0001 1111 1111 1100 0000 0000)
  REG_PIOC_OER =  0x000071FE;     // Set bits PortC to outputs. B(0000 0000 0000 0000 0111 0001 1111 1110)

  // Not doing Port D (Ram Data Bus). Data direction set in ram_mode() function.

  // Set all CTRL bits high 
  set_ctrl_bits(0x00007000);     // B(0000 0000 0000 0000 0111 0000 0000 0000). CE OE WE
  
  //--------------------------------------------------------------------------------------------------------------------------------------------------------------------

  // Change A/D resolution to 12 bits
  analogReadResolution(12);
  
	size_of_sample_data = write_wav_data();	// Get Sample from card and read into RAM.
	
	// Check for minimum filesize, 4096 samples, 8192 bytes
	if(size_of_sample_data < 8192)
	{
		Serial.print("File size too small, must be > 8192 + 44 byte header. 8236 bytes minimum"); 
		fast_flash();	// Signal filesize too small using a fast flash
	}
	
	// Create Segment and Zero Crossing Tables
	make_spl_seg_tbl(); // Dont need to pass tables as they are declared as global
	find_zero_xings();		

	// Check Trigger edge select.
	trig_pol = digitalRead(69);	// Read trig polarity input 
	
	if(trig_pol == LOW)	// Link on, -ve edge trig
	{
		Serial.println("Negative edge trigger selected");
	}
	else
	{
		Serial.println("Positive edge trigger selected");
	}

  // ZXD switch D64. link off = off, link on = on.
  if(digitalRead(64) == 1)
  {
    // Zero Crossing off
    Serial.println("Zero Crossing Detection Disabled");  
  }
  
  else 
  {
    // Zero Crossing on
    Serial.println("Zero Crossing Detection Enabled");
  }

  
	// Do a set of Switch and A/D reads before starting playback.
	do_switch_reads();
	do_atod_reads();

	abs_spl_read_add = playback_start_add;		// assign this for first time round loop.
	
	Serial.println("\nPlaying Sample");
	start_playback();	// enables playback interupt. 
}

// --------------- Main Loop ----------------
void loop() 
{
	// Do Switch and A/D reads.		
	do_atod_reads();
	do_switch_reads();	
	
	// Check for Retrig Flag
	// If set, reset to start of loop.
	// Reset Retrig flag 
	
	checktrig(); // check trigger
	
	if(mode_sw == 0 && has_triggered == true)	// one shot enabled and trigger has been detected
	{
		// Restart loop
		if(fwd_rev_sw == 1)		// check direction of playback and reset accordingly
		{
			abs_spl_read_add = playback_end_add;
			has_triggered = false;		// Reset Trigger flag
			pb_direction = 1;	// set interupt playback direction
		}
		else
		{
			abs_spl_read_add = playback_start_add;
			has_triggered = false;		// Reset Trigger flag
			pb_direction = 0;	// set interupt playback direction
		}
		//turn on interupt
		NVIC_EnableIRQ(TC0_IRQn);    
	}
	
	if(mode_sw == 1)	// Loop enabled
	{   	
		//turn on interupt
		NVIC_EnableIRQ(TC0_IRQn);
		
		if(restart_sw == 0 && has_triggered == true)	// Check for sample restart
		{
			if(fwd_rev_sw == 1)		// check direction of playback and reset accordingly
			{
				abs_spl_read_add = playback_end_add;
				has_triggered = false;		// Reset Trigger flag
			}
			else
			{
				abs_spl_read_add = playback_start_add;
				has_triggered = false;		// Reset Trigger flag
			}
		}
	}

	delay(1); // Need this delay otherwise the interupt has problems. Dont know why	
}

// ----------------------------------------------- Function definitions -----------------------------------------------------

void checktrig()
{
	// Check for +ve or -ve going trigger depending on trig polarity input read on startup.
	// need to reset has_triggered in calling routine
	
	static bool prev_trig_state;	// preserve value between calls
	
	bool trig_state = digitalRead(46);	// Read trig input
	
	if(trig_pol == LOW)	// Link on, Check for -ve edge trig
	{
		if (trig_state == LOW && prev_trig_state == HIGH)
		{
			has_triggered = true;
		}
		else
		{
			has_triggered = false;
		}
	}
	
	else	// Link off, Check for +ve going edge on trigger input.
	{
		if (trig_state == HIGH && prev_trig_state == LOW)
		{
			has_triggered = true;
		}
		else
		{
			has_triggered = false;
		}
	}
	prev_trig_state = trig_state;
	return;
}

//--------------------------------------------------------------------------------------------------------------------------------

void slow_flash()	// Flashes slowly, endless loop. Indicates SD init fail
{
	while(1)
	{
		delay(1000);
		digitalWrite(LED_BUILTIN, !digitalRead(LED_BUILTIN));
	}
}

//--------------------------------------------------------------------------------------------------------------------------------
		
void fast_flash()	// Flashes more quickly, endless loop. Indicates Sample file not found or file size too small.
{ 
	while(1)
	{
		delay(100);
		digitalWrite(LED_BUILTIN, !digitalRead(LED_BUILTIN));
	}
}
 
//--------------------------------------------------------------------------------------------------------------------------------

// Write to Ext Ram address bus. Send 19 bit address
void set_ram_add_bits(unsigned long add_bits)
{
    unsigned long shifted_bits;   // Storage for first shift
    
    // Lower 8 bits (0 - 7) to PC01 - PC08
    // Upper 11 bits (8 - 18) to PA10 - PA20

    // PIO and OER Registers must be set before calling.

    // Mask writes to address bits only. Must write to Enable and disable registers (OWER and OWDR). One is inverse of other.
    // Port A Mask
    REG_PIOA_OWER = 0x001FFC00;     // Enable writes to Port A bits. B(0000 0000 0001 1111 1111 1100 0000 0000)
    REG_PIOA_OWDR = 0xFFE003FF;     // Disables writes to Port A bits B(1111 1111 1110 0000 0000 0011 1111 1111)
    
    // Port C Mask
    REG_PIOC_OWER = 0x000001FE;     // Enable writes to Port C bits. B(0000 0000 0000 0000 0000 0001 1111 1110)
    REG_PIOC_OWDR = 0xFFFFFE01;     // Disableswrites to Port C bits B(1111 1111 1111 1111 1111 1110 0000 0001)

    // Unused port bits are masked so we should only need to shift long_index right by 1 and send to Port C for lower 8 bits
    // then shift again and send to Port A for upper 11 bits 

    // Write to bits PC01 - PC08.
    shifted_bits = (add_bits << 1); // First shift A0 = A1 etc.
    REG_PIOC_ODSR = shifted_bits;  // Write to Port C 

    // Write to bits PA10 - PA20.
    REG_PIOA_ODSR = (shifted_bits << 1);  // Another shift and write A10 = A11 etc. 

    return;
}

//--------------------------------------------------------------------------------------------------------------------------------

// Write to Ext Ram data bus. 8 bits.
void set_ram_dat_bits(unsigned long data_bits)
{
  // 8 bits (0 - 7) to PD00 - PD07. Unused upper bits masked.

  // PIO and OER Registers must be set before calling.

  // Mask writes to address bits only. Must write to Enable and disable registers (OWER and OWDR). One is inverse of other.
  // Port D Mask
  REG_PIOD_OWER = 0x000000FF;     // Enable writes to Port D bits. B(0000 0000 0000 0000 0000 0000 1111 1111)
  REG_PIOD_OWDR = 0xFFFFFF00;     // Disables writes to Port D bits B(1111 1111 1111 1111 1111 1111 0000 0000)
    
  // Write to bits PC01 - PC08.
  REG_PIOD_ODSR = (data_bits);  // Write to Port D  

  return;
}

//---------------------------------------------------------------------------------------------------------------------------------

// External Ram Read
// Ram mode set to Read at end of write_wav_data() function.

byte read_ex_ram(unsigned long read_add)
{
  unsigned long value_read;
  
  set_ram_add_bits(read_add);    // Present Address
  set_ctrl_bits(0x00003000);     // B(0000 0000 0000 0000 0011 0000 0000 0000). NCE OE WE
  set_ctrl_bits(0x00001000);     // B(0000 0000 0000 0000 0001 0000 0000 0000). NCE NOE WE
  
  value_read = REG_PIOD_PDSR;    // Read Port D
  
  set_ctrl_bits(0x00005000);     // B(0000 0000 0000 0000 0101 0000 0000 0000). CE NOE WE
  set_ctrl_bits(0x00007000);     // B(0000 0000 0000 0000 0111 0000 0000 0000). CE OE WE
  
  return byte(value_read);       // Cast to byte and return
}

//----------------------------------------------------------------------------------------------------------------------------------

// External Ram Write 
void write_ex_ram(unsigned long write_add, byte write_dat)
{
  // Set Ram Write mode
  ram_mode('W');

  set_ram_add_bits(write_add);    // Present Address
  set_ram_dat_bits(write_dat);    // Present Data
  set_ctrl_bits(0x00003000);     // B(0000 0000 0000 0000 0011 0000 0000 0000). NCE OE WE
  set_ctrl_bits(0x00002000);     // B(0000 0000 0000 0000 0010 0000 0000 0000). NCE OE NWE
  set_ctrl_bits(0x00003000);     // B(0000 0000 0000 0000 0011 0000 0000 0000). NCE OE WE
  set_ctrl_bits(0x00007000);     // B(0000 0000 0000 0000 0111 0000 0000 0000). CE OE WE
  return;
}

//---------------------------------------------------------------------------------------------------------------------------------

// Ext Ram Mode (Set Port D to input or output (Ram Data Bus).
void ram_mode(char mode)
{
  switch (mode)
  {
    case 'R':
      // do read mode stuff
      REG_PIOD_OER = 0x00000000;     // Set the lowest 8-bits of Port D to inputs. B(0000 0000 0000 0000 0000 0000 0000 0000)
      REG_PIOD_ODR = 0x000000FF;     // Disable write to lower 8 bit of Port D.     B(0000 0000 0000 0000 0000 0000 1111 1111)
    break;
    case 'W':
      // do write mode stuff
      REG_PIOD_OER =  0x000000FF;     // Set the lowest 8-bits of PortD to outputs. B(0000 0000 0000 0000 0000 0000 1111 1111)
    break;
  }
  return;
}

//---------------------------------------------------------------------------------------------------------------------------------

// Write to Ext Ram CTRL bits. Doing this the long way to gain speed.
// NCE = PC14, NOE = PC13, NWE = PC12
void set_ctrl_bits(unsigned long ctrl_bits)
{
    // Port C Mask
    REG_PIOC_OWER = 0x00007000;     // Enable writes to Port C Ram CTRL bits. B(0000 0000 0000 0000 0111 0000 0000 0000)
    REG_PIOC_OWDR = 0xFFFF8FFF;     // Disable writes to Port C RAM CTRL bits. B(1111 1111 1111 1111 1000 1111 1111 1111)
    
    REG_PIOC_ODSR = ctrl_bits;  // Do the write
    return;
}

//---------------------------------------------------------------------------------------------------------------------------------

// Transfer WAV data from SD Card to External Ram.
// WAV file filename : sample.wav
// Not going to bother with examining the WAV header as
// the Sketch is written to accept WAV data 16 Bit signed mono only.
// The user is free to try other formats and find out what happens.

// Read data starting from 45th Byte in file (location 44).
// Read two bytes at a time. 
// WAV format is little endian so we need to swap bytes.

// SD card attached to SPI bus as follows:
// MOSI - pin 11
// MISO - pin 12
// CLK - pin 13
// CS - pin 4

// Returns size of data in Bytes. Size 0 indicates an error.

int write_wav_data(void)
{
  File myFile;
  int ext_ram_add = 0;	// Ram address counter, 16 bit variable pointing to Byte in ext ram.
  byte spl_lsb; // sample byte read from SD lsb
  byte spl_msb; // sample byte read from SD msb
  
  digitalWrite(LED_BUILTIN, LOW);		// Turn on Loading / Error LED to signal sample being loaded.

  Serial.print("Initializing SD card...");

  // Pin 10 must be left as an output or the SD library functions will not work. 
  pinMode(10, OUTPUT);

  if (!SD.begin(4)) 
  {
    Serial.println("initialization failed!");
    
    // Slow flash Builtin LED to signal SD Card init fail.
    slow_flash();

  }
  Serial.println("initialization done.");
  
  // open sample.wav on sd card for reading and save to ext ram
  myFile = SD.open("sample.wav");
  if (myFile) 
  {
	Serial.print("File size on SD Card: ");
	Serial.print(myFile.size());
	Serial.println(" Bytes");
    Serial.println("opening and writing sample.wav to external RAM");
    
      // Skip Header (44 Bytes). Not needed as format specified.
	  for(int skip = 0; skip < 44; skip++)
	  {
		  myFile.read();
	  }

    // read from the file until there's nothing else in it:
    while (myFile.available()) 
    { 
      // Do sample manipulation stuff here
      spl_lsb = byte((myFile.read()));  // Little endian
      spl_msb = byte((myFile.read()));
      
      // Combine two bytes together using 16bit = (8bit_msb * 256) + 8bit_lsb.
      uint16_t signed_sixteenbit_spl = (spl_msb * 0x100) + spl_lsb; // Unsign as we want to leave the value as is.
	 
      // Convert to unsigned by inverting MSB. Exor with 0x8000.
      uint16_t unsigned_sixteenbit_spl = signed_sixteenbit_spl ^ 0x8000; 

      // Convert 16 bit samples to 12 bit for use on Due DAC. 
      // Do it by shifting bits right
	  uint16_t twelvebit_spl = unsigned_sixteenbit_spl >> 4; 

      // split 12 bit sample into msb and lsb and store to external Ram. Little endian.
      byte spl_lo = twelvebit_spl & 0x00FF; 		// And lower bits
	  byte spl_hi = (twelvebit_spl & 0x0F00) >> 8;		// And upper bits, shift into lower byte 
  
      // Store to external ram
      write_ex_ram(ext_ram_add, spl_lo);
      write_ex_ram(ext_ram_add+1, spl_hi);
      
      ext_ram_add += 2;  // inc ram address by two bytes (one sample)
      
      // Check for Ram full. Just stop reading and writing.
      if (ext_ram_add > 0x80000)		// 0x80000 = 512 * 1024 Dec. 
      {
		  break;	
	  }
    }
    // close the file and store size
    myFile.close();
    
    Serial.println("External Ram write completed.");
    Serial.print("Size of sample on External RAM = ");
    Serial.print(ext_ram_add);
    Serial.println(" Bytes.");
    
  } 
  
  else 
  {
    // if the file didn't open, print an error:
    Serial.println("Error opening sample.wav.");
    
    // Fast flash Builtin LED to signal File not found.
	fast_flash();
  }
  digitalWrite(LED_BUILTIN, HIGH);		// Turn off Loading / Error LED to signal Sample loaded.
  
  ram_mode('R');	// Finished writing so can set Ext Ram to read mode
  
  return ext_ram_add;	// Contains number of bytes
}

//---------------------------------------------------------------------------------------------------------------------------------

void start_playback(void)
{
	// Interupt setup. Starts Sample playing

	// Timer Counter 0 Channel 0 to generate pulses TIOA0 
	PMC->PMC_PCER0 |= PMC_PCER0_PID27;		// TC0 power ON - Timer Counter 0 channel 0 IS TC0

	PIOB->PIO_PDR |= PIO_PDR_P25;				// The pin is no more driven by the GPIO
	PIOB->PIO_ABSR |= PIO_PB25B_TIOA0;			// TIOA0 (pin 2) is PB25 peripheral type B

	TC0->TC_CHANNEL[0].TC_CMR = TC_CMR_TCCLKS_TIMER_CLOCK1  // MCK/2, clk on rising edge
								| TC_CMR_WAVE               		// Waveform mode
								| TC_CMR_WAVSEL_UP_RC        // UP mode with automatic trigger on RC Compare
								| TC_CMR_ACPA_CLEAR          // Clear TIOA0 on RA compare match
								| TC_CMR_ACPC_SET;           // Set TIOA0 on RC compare match

	TC0->TC_CHANNEL[0].TC_RA = 5;			// Duty cycle = (TC_RA/TC_RC) X 100%

	TC0->TC_CHANNEL[0].TC_IER = TC_IER_CPCS;		// Interrupt on RC comp match
	NVIC_EnableIRQ(TC0_IRQn);                            // TC1 Interrupt enable

	TC0->TC_CHANNEL[0].TC_CCR = TC_CCR_SWTRG | TC_CCR_CLKEN; // Trig TC0 counter and enable
	
	return;
}

//---------------------------------------------------------------------------------------------------------------------------------

void do_atod_reads()	// Reads A/D values, start uses segment or zero crossing table to calculate addresses. Data returned to global variables.
// To switch between ZXD and SPL tables we should only need to reassign the pointer below and set the A/D values
{
	int num_of_atod_reads = 5;	// number of A/D reads to average out
	int dummy_read;	
	int AD_val_min_diff = 5;	// Only update A/D values if over this value. Helps eliminate jitter. 
	
	unsigned int start_cv_sum = 0; 
	unsigned int length_cv_sum = 0;
	unsigned int rate_cv_sum = 0;
	unsigned int start_off_sum = 0;
	unsigned int length_off_sum = 0;
	unsigned int rate_off_sum = 0;
	
	// Bloody pointers are doing my head in. Thanks Stack Exchange for helping.
	unsigned int *tbl_ptr;	// init table pointer
	
	// ZXD switch D64. link off = off, link on = on.
	if(digitalRead(64) == 1)
	{
		// Zero Crossing off, assign Table Pointer to sample table
		tbl_ptr = &spl_add_tbl[0];	
	}
	
	else 
	{
		// Zero Crossing on, assign Table pointer to Zero Crossing Table and set A/D read mapping.
		tbl_ptr = &zxd_add_tbl[0];	
	}
	
	// Read and average all analogue inputs at once
	// Do dummy reads first to eliminate crosstalk
	for(int count = 0; count < num_of_atod_reads; count ++)
	{
		dummy_read = analogRead(A6);
		start_cv_sum += analogRead(A6);
		
		dummy_read = analogRead(A5);
		length_cv_sum += analogRead(A5);
		
		dummy_read = analogRead(A7);
		rate_cv_sum += analogRead(A7);
		
		dummy_read = analogRead(A3);
		start_off_sum += analogRead(A3);
		
		dummy_read = analogRead(A2);
		length_off_sum += analogRead(A2);
		
		dummy_read = analogRead(A4);
		rate_off_sum += analogRead(A4);
	}
	
	// Calculate the mean values
	int start_cv_avg_value = start_cv_sum / num_of_atod_reads;
	int length_cv_avg_value = length_cv_sum / num_of_atod_reads;
	int rate_cv_avg_value = rate_cv_sum / num_of_atod_reads;
	int start_off_avg_value = start_off_sum / num_of_atod_reads;
	int length_off_avg_value = length_off_sum / num_of_atod_reads;
	int rate_off_avg_value = rate_off_sum / num_of_atod_reads;
	
	// -------------------------------- Sample playback rate value manipulation --------------------------------

	// The values of the Rate CV input and Offset are added together. and constrained to a maximum of 4095
	int rate_value = constrain(rate_cv_avg_value + rate_off_avg_value, 0, 4095);
	
	// Highest Playback rate value is 250. any lower (Higher Frequency) and operation is erratic.
	int mapped_value = map(rate_value, 0, 4095, 1950, 3845);	

	TC0->TC_CHANNEL[0].TC_RC = 4095 - mapped_value;  		// Frequency = (Mck/2)/TC_RC. invert so that pot works the right way round
	// The next line needs to be here or else the interupt stops if the control signals change to quickly.
	TC0->TC_CHANNEL[0].TC_CCR = TC_CCR_SWTRG | TC_CCR_CLKEN; // Trig TC0 counter and enable 
	
	// -------------------------------------------------------------------------------------------------
	
	// The values of the Length CV input and Offset are added together. and constrained to a maximum of 4095 segments
	int length_value = constrain(length_cv_avg_value + length_off_avg_value, 0, 4095);

	// The values of the Start CV input and Offset are added together. and constrained to a maximum of 4095 segments
	int start_segment = constrain(start_cv_avg_value + start_off_avg_value, 0, 4095);
	
	int difference = abs(start_segment - org_start_pos);	// Check that start segment number has changed by a certain amount to eliminate A/D jitter
	
	if(difference > AD_val_min_diff )
	{
		org_start_pos = start_segment;	// Update 
	}

	int end_segment = start_segment + length_value;	// Calculate end segment.	
	
	difference = abs(end_segment - org_end_pos);	// Check that end segment number has changed by a certain amount to eliminate A/D jitter
	
	if(difference > AD_val_min_diff)
	{
		org_end_pos = end_segment;	// Update 
	}
	
	// Check that loop end segment does not overun. (Segment number not greater than 4095)
	org_end_pos = constrain(org_end_pos, 0, 4095);
	
	// get playback start address from whichever address table we are using
	playback_start_add = *(tbl_ptr + org_start_pos);	
	playback_end_add = *(tbl_ptr + org_end_pos);
	
	return;
}

//---------------------------------------------------------------------------------------------------------------------------------

void do_switch_reads(void)
{
	bounce_sw = digitalRead(62);	// Bounce switch is actioned in interupt handler
	fwd_rev_sw = digitalRead(63);	//
	
	if(bounce_sw == 1)	// Only assign direction if bounce is off
	{
		pb_direction = fwd_rev_sw; 
	}
	
	mode_sw = digitalRead(48);	// Loop / One-Shot 
	restart_sw = digitalRead(65);	// restart playback. This will also retrigger the playback in One-Shot mode
	
	return;
}

//---------------------------------------------------------------------------------------------------------------------------------

void TC0_Handler()	 
{
	if(pb_direction == 1)	// FWD = 0, REV = 1
	{	// Reverse
		abs_spl_read_add -= 2;	// Dec read address
		
		if(abs_spl_read_add <= playback_start_add)	// Start of Sample ?
		{			
			if(bounce_sw == 0)	// Bounce on ?
			{
				pb_direction = ! pb_direction;	// Toggle direction at end of loop if bounce enabled. pb_direction assignment from Fwd/Bwd switch surpressed 
				
				if(mode_sw == 0)	// One shot selected ?
				{
					if(fwd_rev_sw == 0)	// Stop interupt here if FWD selected. Does FWD, then this REV, and stops at end of playback
					{
						// turn off interupt
						NVIC_DisableIRQ(TC0_IRQn);
					}
				}
			}
			else
			{
				if(mode_sw == 0)// one shot enabled ?
				{
					// turn off interupt
					NVIC_DisableIRQ(TC0_IRQn);
				}
				
				abs_spl_read_add = playback_end_add;	// only reset to start position if bounce off
			}
		}			
	}

	else	
	{	// Forward
		abs_spl_read_add += 2;	// Inc read address
		
		if(abs_spl_read_add >= playback_end_add)	// End of sample ?
		{
			if(bounce_sw == 0)	// Bounce on ? 
			{
				pb_direction = ! pb_direction;	// Toggle direction at end of loop if bounce enabled.  pb_direction assignment from Fwd/Bwd switch surpressed
				
				if(mode_sw == 0)	// One shot selected ?
				{
					if(fwd_rev_sw == 1)	// Stop interupt here if REV selected. Does REV, then this FWD, and stops at end of playback
					{
						// turn off interupt
						NVIC_DisableIRQ(TC0_IRQn);
					}
				}
			}	
			else
			{
				if(mode_sw == 0)// one shot enabled ?
				{
					// turn off interupt
					NVIC_DisableIRQ(TC0_IRQn);
				}
				
				abs_spl_read_add = playback_start_add;	// only reset to start position if bounce is off.
			}			
		}
	}
		
	TC0->TC_CHANNEL[0].TC_SR;  // Read and clear status register
	
	byte spl_read_lo = read_ex_ram(abs_spl_read_add);	// Lo byte
	byte spl_read_hi = read_ex_ram(abs_spl_read_add + 1);	// Hi byte
	
	// Combine two bytes together using 16bit = (8bit_msb * 256) + 8bit_lsb.
	int16_t sample_word = (spl_read_hi * 0x100) + spl_read_lo;

	/* write to DAC0 */
	dacc_write_conversion_data(DACC_INTERFACE, sample_word);
	
	return; 
}

//---------------------------------------------------------------------------------------------------------------------------------

void make_spl_seg_tbl()
{
	// Calculate segment size in samples (2 Bytes).
	unsigned int num_of_samples = size_of_sample_data / 2;	// samples are two bytes each. size of sample data is in single bytes.
	
	sample_segment_size = num_of_samples / 4096;	// Size of each sample segment in samples (Bytes X 2).
	
	unsigned int bytes_segment_size = sample_segment_size * 2;
	
	Serial.print("\nSample Segment Size: ");
	Serial.print(sample_segment_size);
	Serial.println(" Samples");
	
	
	unsigned int *spl_tbl_ptr = &spl_add_tbl[0];	// assign pointer to first array element sample segment address table

	for(unsigned int index = 0; index < 4096; index ++)	
	{
		*(spl_tbl_ptr + index) = sample_segment_size * index * 2;	// calculate and store each segment start address in array (Bytes not samples)	
	}

	Serial.println("Sample Segment Table created.");
	
	return;
}	

//---------------------------------------------------------------------------------------------------------------------------------
	
void find_zero_xings()
{
	// Now lets find those pesky Zero Crossings
	
	// Doing this by scanning each sample segment and finding the position of the first sample lowest to Zero.
	// using unsigned data which means that the zero point is at a value of 2048.
	
	unsigned int *zxd_tbl_ptr = &zxd_add_tbl[0];	// assign pointer to first array element zero crossing address table
	
	// Calculate segment size in samples (2 Bytes).
	unsigned int num_of_samples = size_of_sample_data / 2;	// samples are two bytes each. size of sample data is in single bytes.
	
	sample_segment_size = num_of_samples / 4096;	// Size of each sample segment in samples (Bytes X 2).
	
	unsigned int bytes_segment_size = sample_segment_size * 2;
	
	// This is the actual part where we find the zero crossings. Addresses in bytes. 
		
	for(unsigned int segment_num = 0; segment_num < 4096; segment_num ++)	// scan through segments
	{
		uint16_t min_dist_val = 2048;	// Value of the nearest sample to zero. Start with highest as this will only decrease.
		
		// make absolute address 
		unsigned int abs_address = *(spl_add_tbl + segment_num);	// Steps through each entry in table of Ext Ram Addresses (Start of Each Segment)
			
		// --------------- Segment scanning loop --------------------- 
		for(unsigned int pos_in_segment = 0; pos_in_segment < bytes_segment_size; pos_in_segment += 2)	// iterate through segment a sample at a time (16 bits)
		{
			// make absolute address
			abs_address = (*(spl_add_tbl + segment_num)) + pos_in_segment;
			
			// Get sample
			byte spl_lsb = read_ex_ram(abs_address);	// lsb
			byte spl_msb = read_ex_ram(abs_address + 1);	// msb
			uint16_t unsigneddat = (spl_msb * 0x100) + spl_lsb;	// Comine bytes into word

		 	// Calculate the distance from zero of sample at this address
		 	int current_dist_from_zero_signed = 2048 - unsigneddat;		// Can't use the ABS function with anything other than a single variable
		 	uint16_t current_dist_from_zero_unsigned = abs(current_dist_from_zero_signed);	// Loose that pesky sign.
		 	
		 	if(current_dist_from_zero_unsigned < min_dist_val)
		 	{
				*(zxd_tbl_ptr + segment_num) = abs_address;	// Current sample value is nearer to Zero than previous closest, store its value in the table
				min_dist_val = current_dist_from_zero_unsigned;	// Store new value of nearest to zero.
				
			}
		}	
	}
	Serial.println("Zero Crossing Table created.\n");
	
	return;	
}	

//--------------------------------------------------------------------------------------------------------------------------------
